import bisect
import json
import random
from . import scheduler
from . import subtasks

TASK_DEBUG = False

class BadRequest(Exception): pass

"""
state = {
    "opened": False,
    "tiers": [
        ["main", 123],  # start time
    ],
    "last_update": 12345,
    "reputation": 5,
    "active_customers": [
        { "tier": "main", "customer_id": 123, "internal_id": 123, "type": "identity", "var1": "...", "answer": "...", "expires": 123 },
    ],
    "won_easy": False,
    "won_hard": False,
    "died": None or 123,
    "next_customer_id": 123,
    "messages": [[123, "..."], [456, "..."]],
    "last_submit_customer_id": 123,  # only after submit
    "last_submit_answer": "...",  # only after submit
}
"""

def get_default_state():
    return {
        "opened": False,
        "tiers": [],
        "last_update": 0,
        "reputation": 10,
        "active_customers": [],
        "won_easy": False,
        "won_hard": False,
        "died": None,
        "next_customer_id": 1,
        "messages": [],
        "wrong_answers": [],
    }

def add_message(state, when, message):
    state["messages"].append([when, message])

def decrease_reputation(state, when):
    state["reputation"] -= 1
    if state["reputation"] == 0:
        add_message(state, when, "Your reputation decreased to 0 and you went bankrupt!")
        state["opened"] = False
        state["died"] = 20*60 - (when - state["tiers"][0][1])
        state["active_customers"] = []
        state["tiers"] = []
        return True
    return False


class TierCursor:
    def __init__(self, state, i):
        self.i = i
        self.tier_id, self.start_time = state["tiers"][i]
        self.events = scheduler.tiers[self.tier_id]["events"]
        times = scheduler.tiers[self.tier_id]["times"]
        self.position = bisect.bisect_right(times, state["last_update"] - self.start_time)

    def next_event(self):
        if self.position == len(self.events): return (True, None, None, None, None, None, None)
        when, event_type, internal_id, types, expires = self.events[self.position]
        when += self.start_time
        if expires is not None: expires += self.start_time
        return (False, when, event_type, self.i, internal_id, types, expires)

    def advance(self):
        self.position += 1


def advance(state, contest_time):
    active_set = { (customer["tier"], customer["internal_id"]) for customer in state["active_customers"] }

    cursors = [TierCursor(state, i) for i in range(len(state["tiers"]))]

    while True:
        if not cursors: break
        finished, when, event_type, tier_index, internal_id, types, expires = min(cursor.next_event() for cursor in cursors)
        if finished: break
        if when > contest_time: break
        cursors[tier_index].advance()
        tier_id, start_time = state["tiers"][tier_index]

        if event_type == scheduler.ENTER:
            type = random.choice(types)
            if type == 'ipschuman':
                customer = subtasks.tasks[type](state)
            else:
                customer = subtasks.tasks[type]()
            customer["tier"] = tier_id
            customer["customer_id"] = state["next_customer_id"]
            customer["internal_id"] = internal_id
            customer["type"] = type
            customer["expires"] = expires
            state["active_customers"].append(customer)
            active_set.add((tier_id, internal_id))
            state["next_customer_id"] += 1
        if event_type == scheduler.LEAVE:
            if (tier_id, internal_id) in active_set:
                add_message(state, when, "You failed to serve a customer in time. Everyone left! (-1 reputation)")
                state["active_customers"] = []
                active_set = set()
                if decrease_reputation(state, when):
                    break
        if event_type == scheduler.WIN_EASY:
            if not state["won_easy"]:
                state["won_easy"] = True
                add_message(state, when, "Congratulations, you have solved subproblem M1!")
        if event_type == scheduler.WIN_HARD:
            if not state["won_hard"]:
                state["won_hard"] = True
                state["active_customers"] = []
                state["tiers"] = []
                add_message(state, when, "Congratulations, you have solved subproblem M2!")

    state["last_update"] = contest_time

def get_next_refresh(state, contest_time):
    next_refresh = contest_time + (10 if state["active_customers"] else 30)
    active_set = { (customer["tier"], customer["internal_id"]) for customer in state["active_customers"] }

    cursors = [TierCursor(state, i) for i in range(len(state["tiers"]))]

    while True:
        if not cursors: break
        finished, when, event_type, tier_index, internal_id, types, expires = min(cursor.next_event() for cursor in cursors)
        if finished: break
        # when > contest_time is allowed
        cursors[tier_index].advance()
        tier_id, start_time = state["tiers"][tier_index]

        if event_type == scheduler.LEAVE and (tier_id, internal_id) not in active_set:
            continue

        next_refresh = min(next_refresh, when)
        break

    return next_refresh

def handle_open(state, req, contest_time):
    if state["opened"]:
        return "Error: already open."

    state["opened"] = True

    if state["died"] is None:
        state["tiers"] = [["main", contest_time]]
        add_message(state, contest_time, "McDroid's is open for business!")
    else:
        state["tiers"] = [["main", contest_time + state["died"]], ["humans", contest_time]]
        state["died"] = None
        state["reputation"] = 5
#        if contest_time < 60*60*2+30*60:
#            state["tiers"] = state["tiers"][0:1]
#            state["reputation"] = 10
        add_message(state, contest_time, "McDroid's is back in business!")

def handle_refresh(state, req, contest_time):
    pass

def handle_submit(state, req, contest_time):
    try:
        customer_id = int(req["customer_id"])
    except Exception:
        return "Error: missing or invalid 'customer_id'."

    answer = req.get("answer")
    if not answer:
        return "Error: missing 'answer'."
    if not isinstance(answer, str):
        return "Error: 'answer' is not a string."

    state["last_submit_customer_id"] = customer_id
    state["last_submit_answer"] = answer[0:100]

    customers = state["active_customers"]
    for i in range(len(customers)):
        if customers[i]["customer_id"] == customer_id:
            customer = customers.pop(i)
            checker = subtasks.checkers.get(customer["type"], subtasks.check_default)
            if not checker(customer["answer"], answer):
                if TASK_DEBUG: print(customer["answer"], answer)
                state["wrong_answers"].append([customer["type"], customer["answer"], answer, customer.get("var1", None), customer.get("var2", None)])
                add_message(state, contest_time, "A customer received the wrong food! (-1 reputation)")
                decrease_reputation(state, contest_time)
                return "Wrong food! The customer is disgusted!"
            return "Well done!"
    return "This customer is already gone." if customer_id < state["next_customer_id"] else "Error: no such 'customer_id'."

handlers = {
    "open": handle_open,
    "refresh": handle_refresh,
    "submit": handle_submit,
}

def process_internal(state, req, real_time, contest_time):
    action = req.get("action")
    if action not in handlers:
        return { "error": "Bad request - missing or invalid 'action'." }

    next_event = advance(state, contest_time)

    result = handlers[action](state, req, contest_time)

    active_customers = []
    for customer in state["active_customers"]:
        public_customer = { k: v for k, v in customer.items() if k in ("customer_id", "var1", "var2", "expires") }
        if TASK_DEBUG: public_customer["_answer"] = customer["answer"]
        public_customer["text"] = subtasks.texts[customer["type"]]
        active_customers.append(public_customer)

    state["messages"] = state["messages"][-10:]

    # return { "error": "You are not logged in.", "login": "http://example.com/" }
    # return { "error": "The contest is over." }

    next_refresh = get_next_refresh(state, contest_time)

    return {
        "real_time": real_time,
        "contest_time": contest_time,
        "result": result,
        "active_customers": active_customers,
        "next_refresh": next_refresh,
        "next_refresh_please_sleep": next_refresh - contest_time,
        "opened": state["opened"],
        "reputation": state["reputation"],
        "messages": state["messages"],
        "won_easy": state["won_easy"],
        "won_hard": state["won_hard"],
    }

def process(state, req, real_time, contest_time):
    state = get_default_state() if state == None or state == '' else json.loads(state)
    won_easy_was = state["won_easy"]
    won_hard_was = state["won_hard"]
    response = process_internal(state, req, real_time, contest_time)
    ok_easy = state["won_easy"] and not won_easy_was
    ok_hard = state["won_hard"] and not won_hard_was
    return response, json.dumps(state, sort_keys=True), ok_easy, ok_hard
